// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use quinn::{Connection, SendDatagramError};
use tokio::io::BufReader;

use super::{MessageStream, MsError, quic_common::{SenderTask, ReceiverTask, ReceiverTaskBuilder, ErrorContainer, MessageSer, MessageDe}};

use crate::net::{serp::Request, message::{Message, WorldMessage, ScoreMessage}};
use crate::player::update::PlayerUpdate;

pub struct QuicClientMs {
	sender: SenderTask<(PlayerUpdate, u64)>,
	receiver: ReceiverTask,
	connection: Connection,
	error: ErrorContainer,
}

impl QuicClientMs {
	pub async fn build(connection: Connection, game_id: Box<str>) -> Result<(QuicClientMs, Message), String> {
		let (sender_stream, receiver_stream) = Request::Play(game_id).send(&connection).await.map_err(|err| format!("failed making request: {err}"))?;
		let score_stream = connection.accept_uni().await.map_err(|err| format!("failed accepting unidirectional score stream: {err}"))?;

		let error = ErrorContainer::default();
		let sender = SenderTask::new(sender_stream, error.clone());
		let mut receiver_builder = ReceiverTaskBuilder::new(error.clone(), 1 << 24).add_stream::<WorldMessage>(receiver_stream);
		let init_msg = receiver_builder.receive(&error, &connection).await.map_err(|err| format!("failed receiving initial message: {err}"))?;
		let receiver = receiver_builder.add_stream::<ScoreMessage>(BufReader::new(score_stream)).build();

		Ok((QuicClientMs { sender, receiver, connection, error }, init_msg))
	}
}

impl MessageStream for QuicClientMs {
	fn send(&mut self, msg: Message) -> Result<(), MsError> {
		match msg {
			Message::PlayerUpdate(update, count) => self.sender.send((update, count), &self.error, &self.connection),
			Message::PlayerBulkUpdate(bulk_update) => {
				let Some(limit) = self.connection.max_datagram_size() else { return Ok(()); };
				let Some(data) = bulk_update.serialise(limit) else { return Ok(()); };

				debug_assert!(data.len() < limit);

				match self.connection.send_datagram(data) {
					Ok(()) | Err(SendDatagramError::UnsupportedByPeer | SendDatagramError::Disabled) => Ok(()), // Failure for some reason isn't an error
					Err(SendDatagramError::TooLarge) => { // Might happen with a TOCTOU so despite the size being limited, I shouldn't consider that an error
						log::warn!("failed sending bulk update datagram, too large");
						Ok(())
					},
					Err(SendDatagramError::ConnectionLost(_)) => Err(MsError::Disconnected(None)),
				}
			},
			_ => unimplemented!(),
		}
	}

	fn receive(&mut self) -> Result<Option<Message>, MsError> {
		self.receiver.receive(&self.error, &self.connection)
	}

	fn flush(&mut self) -> Result<(), MsError> {
		self.sender.flush(&self.error, &self.connection)
	}

	fn is_networked(&self) -> bool {
		true
	}
}

impl MessageSer for (PlayerUpdate, u64) {}

impl MessageDe for WorldMessage {
	fn into_message(self) -> Message {
		Message::World(self)
	}
}

impl MessageDe for ScoreMessage {
	fn into_message(self) -> Message {
		Message::Score(self)
	}
}
