// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
pub mod config;
mod player_update_queue;
mod endpoint;
mod repl;

use std::str::FromStr;
use std::collections::{BTreeMap, BTreeSet};
use std::{fs::{self as std_fs, File}, io::BufWriter};
use std::{mem, num::NonZeroUsize};
use std::net::{IpAddr, SocketAddr};
use std::time::{Instant, Duration};
use std::sync::{Arc, atomic::{AtomicUsize, Ordering}};

use ron::de;
use png::{Encoder, ColorType, BitDepth};
use tokio::{time, fs as tokio_fs, sync::{mpsc::Receiver, oneshot::Sender as OneshotSender}};
use quinn::VarInt;

use config::Config;
use player_update_queue::PlayerUpdateQueue;
use endpoint::{router::{Router, RouterEntry, GameServer}, ServerEndpoint};

use super::message::{Message, WorldMessage, ScoreMessage};
use super::message_stream::{DynMs, MsError};
use super::connection_listener::DynConnLis;
use super::lan_discovery::{LanServerInfo, server::LanDiscoveryServer};

use crate::{ServerArgs, LanDiscoveryConfig};
use crate::world::{World, DMap, changes::ServerChanges, update::WorldUpdate, event::{EventChooser, portal::PortalManager, effect_zone::EffectZones}};
use crate::scores::Scores;
use crate::blocks::{Blocks, compressed::CompressedBlocks};
use crate::utils::{Clock, smallest_unused, ctrl_c::CtrlCHandler};

pub const PORT: u16 = 4080; // UDP
const SANITY_CHECK_PERIOD: f32 = 0.25; // Four per second
pub(super) const SHUTTING_DOWN_MSG: &[u8] = b"server shutting down";
pub(super) const SHUTTING_DOWN_STATUS: VarInt = VarInt::from_u32(0);

pub(super) struct Server {
	clients: BTreeMap<u32, Client>,
	removed: BTreeSet<u32>,

	world: World,
	max_players: usize,
	event_chooser: EventChooser,
	scores: Scores,
	blocks: Blocks,
	compressed_blocks: CompressedBlocks,
	clock: Clock,

	connection_listener: Box<DynConnLis>,

	next_sanity_check: f32,
}

pub(super) struct CertInfo {
	pub cert_path: Box<str>,
	pub key_path: Box<str>,
}

struct Client {
	updates: PlayerUpdateQueue,
	stream: Box<DynMs>,
}

impl Server {
	pub(super) fn build(connection_listener: Box<DynConnLis>, config: Config, sanity_checks: bool) -> Result<Server, String> {
		config.validate().map_err(|err| format!("invalid config: {err}"))?;

		let block_map = config.generate_block_map().map_err(|err| format!("cannot generate block map: {err}"))?;
		let blocks = Blocks::new(config.blocks.block_size, config.blocks.grid_size, block_map.iter().map(|&x| x > 0.0).collect());
		let compressed_blocks = CompressedBlocks::from(&blocks);

		let mut special_areas = config.special_areas.areas;
		let portal_manager = PortalManager::build(config.special_areas.portals, config.special_areas.portal_default_colours, &mut special_areas).map_err(|err| format!("portals: {err}"))?;
		let effect_zones = EffectZones::new(config.special_areas.effect_zones, &config.events, config.special_areas.effect_zone_default_colours, &mut special_areas);

		special_areas.retain(|area| area.colour[3] > 0); // Fully transparent special areas are useless, so don't send them
		let world = World::new(config.world_size, special_areas.into_boxed_slice());

		let event_chooser = EventChooser::build(&world, &blocks, portal_manager, effect_zones, config.events)?;

		Ok(Server {
			clients: BTreeMap::new(),
			removed: BTreeSet::new(),
			world,
			max_players: config.max_players.map_or(usize::MAX, NonZeroUsize::get),
			event_chooser,
			scores: Scores::default(),
			blocks,
			compressed_blocks,
			clock: Clock::new_world(),
			connection_listener,
			next_sanity_check: if sanity_checks { SANITY_CHECK_PERIOD } else { f32::INFINITY },
		})
	}

	async fn run(mut self, mut lan_info: Option<(LanServerInfo, Receiver<OneshotSender<LanServerInfo>>)>, shared_player_count: Arc<AtomicUsize>) {
		const PERIOD: Duration = Duration::from_nanos(10_416_667); // 1/96 seconds

		let mut prev_time = Instant::now();
		let mut prev_player_count = 0;

		log::trace!(target: "perf", "period,{}", PERIOD.as_secs_f64());

		loop {
			let player_count = self.world.player_count();
			shared_player_count.store(player_count, Ordering::Relaxed);

			if let Some((info, receiver)) = &mut lan_info {
				while let Ok(sender) = receiver.try_recv() {
					let _ = sender.send(info.clone().with_cur_players(player_count.try_into().unwrap_or(u16::MAX)));
				}
			}

			let cur_time = Instant::now();
			let dt = cur_time.duration_since(prev_time).as_secs_f32();
			prev_time = cur_time;

			if self.update(dt).is_err() {
				for client in self.clients.into_values() {
					client.stream.close();
				}

				return;
			}

			if prev_player_count != player_count {
				log::info!("player count = {player_count}");
				log::trace!(target: "perf", "player count,{player_count}"); // Slightly uglier format for the performance log
				prev_player_count = player_count;
			}

			let elapsed = cur_time.elapsed();
			let time = elapsed.as_secs_f64();
			log::trace!(target: "perf", "loop,{time}");

			time::sleep(PERIOD.saturating_sub(elapsed)).await;
		}
	}

	/**
	 * Returns Ok(()) if everything's fine, and Err(()) if the server should
	 * close.
	 */
	pub fn update(&mut self, dt: f32) -> Result<(), ()> {
		/*
		 * Need at least one step to prevent a player joining being added to
		 * new_players but no world updates occurring for them to be added to the
		 * world. Then the InitWorld message is sent to the client without
		 * containing the player, which is a violation of the protocol and so the
		 * client leaves the game.
		 */
		let steps = self.clock.advance(dt);
		if steps == 0 { return Ok(()); }

		// 1. Listens for new players and adds them to the World
		let mut messages = Vec::new();
		let mut all_acks = Vec::new();
		let mut new_players = BTreeSet::new();

		{
			let mut player_count = self.world.player_count();
			loop {
				match self.connection_listener.get() {
					Ok(Some(incoming)) => {
						if player_count < self.max_players {
							let stream = match incoming.accept() {
								Ok(stream) => stream,
								Err((err, source)) => {
									log::info!("failed accepting player: {err}, source = {source}");
									continue;
								},
							};

							let id = smallest_unused::find(&self.clients);
							let source = stream.source();
							self.clients.insert(id, Client { updates: PlayerUpdateQueue::new(), stream });
							new_players.insert(id);
							messages.push(self.scores.add(id));
							log::info!("player (id = {id}) joined, source = {source}");

							player_count += 1;
						} else {
							incoming.reject(format!("too many players, limit is {}", self.max_players));
						}
					},
					Ok(None) => break,
					Err(()) => return Err(()),
				}
			}
		}

		// 2. Iterating through client message streams to enqueue received player updates
		self.clients.retain(|&id, client| {
			client.updates.begin_iter();

			loop {
				match client.stream.receive() {
					Ok(Some(Message::PlayerUpdate(update, steps))) => client.updates.push(update, steps),
					Ok(Some(Message::PlayerBulkUpdate(bulk_update))) => client.updates.bulk_update(bulk_update),
					Ok(Some(_)) => {
						self.removed.insert(id); // Violation of the protocol so removed
						log::warn!("player (id = {id}) violated the protocol, sending an invalid message");
						return false;
					}
					Ok(None) => return true,
					Err(err) => {
						if let MsError::Other(err) = err {
							log::warn!("player (id = {id}) disconnected with an error: {err}");
						}
						self.removed.insert(id);
						return false;
					}
				}
			}
		});

		// 3. Removing all players to be removed
		for id in mem::take(&mut self.removed) {
			log::info!("player (id = {id}) left");
			messages.push(self.scores.remove(id));
		}

		for i in 0..steps {
			// 4. Creating the player update map and all acknowledgements
			let (player_updates, acks) = {
				let mut updates = DMap::new();
				let mut acks = DMap::new();

				for (&id, client) in &mut self.clients {
					let (update, ack) = client.updates.pop();
					updates.insert(id, update);
					acks.insert(id, ack);
				}

				(updates, acks)
			};

			// 5. Constructing the pre_events
			let mut changes = ServerChanges::default();
			let pre_events = self.event_chooser.get_pre_events(&self.world, &self.blocks, (i == 0).then_some(&new_players));
			self.world.apply_events(&pre_events, &mut changes);

			// 6. Updating the world
			self.world.update(&player_updates, &self.blocks, &mut changes);

			// 7. Updating the scores
			for &killer in &changes.killers {
				messages.push(self.scores.add(killer));
			}

			// 8. Constructing the post_events
			let post_events = self.event_chooser.get_and_apply_post_events(&mut self.world, &self.blocks, &changes);

			// 9. Constructing the message for this update
			let update_msg = Message::World(WorldMessage::WorldUpdate(WorldUpdate { pre_events, players: player_updates, post_events }, 0));

			messages.push(update_msg);
			all_acks.push(acks);
		}

		// 10. Balances jitter and input lag in the player queues
		for client in self.clients.values_mut() {
			client.updates.after_popped();
		}

		// 11. Add any sanity check messages
		self.next_sanity_check -= dt;
		if self.next_sanity_check <= 0.0 {
			messages.push(Message::World(WorldMessage::SanityCheck(self.world.clone())));
			self.next_sanity_check += SANITY_CHECK_PERIOD;
		}

		// 12. Sends out the messages
		self.clients.retain(|&id, client| {
			let messages = if new_players.contains(&id) {
				&[Message::World(WorldMessage::InitWorld {
					player_id: id,
					world: self.world.clone(),
					blocks: self.compressed_blocks.clone(),
				}), Message::Score(ScoreMessage::InitScores(self.scores.clone()))]
			} else {
				messages.as_slice()
			};

			if let Err(err) = Server::send_messages(messages, client.stream.as_mut(), id, &all_acks) {
				self.removed.insert(id);
				if let MsError::Other(err) = err {
					log::warn!("player (id = {id}) disconnected with an error: {err}");
				}
				false
			} else { true }
		});

		Ok(())
	}

	fn send_messages(messages: &[Message], stream: &mut DynMs, id: u32, acks: &[DMap<u32, u64>]) -> Result<(), MsError> {
		let mut ack_index = 0;
		for msg in messages {
			if let Message::World(WorldMessage::WorldUpdate(update, _)) = &msg {
				stream.send(Message::World(WorldMessage::WorldUpdate(update.clone(), acks[ack_index][&id])))?;
				ack_index += 1;
			} else {
				stream.send(msg.clone())?;
			}
		}
		stream.flush()
	}
}

fn deserialise_config(data: Vec<u8>) -> Result<Config, String> {
	de::from_bytes(&data).map_err(|err| format!("invalid config file: {err}"))
}

fn load_config(path: &str) -> Result<Config, String> {
	deserialise_config(std_fs::read(path).map_err(|err| format!("cannot read file \"{path}\": {err}"))?)
}

async fn load_config_async(path: &str) -> Result<Config, String> {
	deserialise_config(tokio_fs::read(path).await.map_err(|err| format!("cannot read file \"{path}\": {err}"))?)
}

#[tokio::main]
pub async fn run(args: ServerArgs) -> Result<(), String> {
	log::info!("Starting the server!");

	let addr = SocketAddr::new(IpAddr::from_str(&args.addr).map_err(|err| format!("invalid address: {err}"))?, args.port);
	let lan_discovery = if args.lan_discovery != LanDiscoveryConfig::Disabled {
		match LanDiscoveryServer::build(addr.ip()) {
			Ok(ld) => {
				log::info!("Created LAN discovery server");
				Some(ld)
			},
			Err(err) => {
				log::warn!("failed creating LAN discovery: {err}");
				None
			},
		}
	} else {
		None
	};

	let mut router = Router::default();
	for game in &args.game_servers {
		let entry = RouterEntry::build(game).map_err(|err| format!("failed parsing router entry: {err}"))?;
		let server = GameServer::build(addr, &entry, args.sanity_checks, lan_discovery.is_some()).await.map_err(|err| format!("failed creating game server: {err}"))?;
		router.add_game_server(entry, server).map_err(|err| format!("failed adding game server: {err}"))?;
	}

	tokio::spawn(async { CtrlCHandler::new().listen().await; });

	let cert_info = args.cert_path.as_ref().zip(args.key_path.as_ref()).map(|c| CertInfo { cert_path: c.0.clone(), key_path: c.1.clone() });
	let endpoint = ServerEndpoint::build(addr, cert_info, router, lan_discovery, args.lan_discovery).await.map_err(|err| format!("failed creating server endpoint: {err}"))?;
	endpoint.run(args, addr).await;

	Ok(())
}

pub fn save_blocks(config_path: &str, output_path: &str) -> Result<(), String> {
	let config = load_config(config_path)?;
	let bmap = config.generate_block_map().map_err(|err| format!("cannot generate config: {err}"))?;
	let file = File::create(output_path).map_err(|err| format!("cannot open \"{output_path}\" for saving image: {err}"))?;

	let mut encoder = Encoder::new(BufWriter::new(file), config.blocks.grid_size.x, config.blocks.grid_size.y);
	encoder.set_color(ColorType::Grayscale);
	encoder.set_depth(BitDepth::Eight);

	let mut writer = encoder.write_header().map_err(|err| format!("cannot write header of image: {err}"))?;
	let mut output_data = Vec::with_capacity(config.blocks.grid_size.element_product() as usize);

	for y in (0..config.blocks.grid_size.y).rev() {
		let i = (y * config.blocks.grid_size.x) as usize;
		for &x in &bmap[i..i + config.blocks.grid_size.x as usize] {
			output_data.push(if x > 0.0 { 0xff } else { 0x00 });
		}
	}

	writer.write_image_data(&output_data).map_err(|err| format!("cannot save image to \"{output_path}\": {err}"))?;
	writer.finish().map_err(|err| format!("cannot flush image data: {err}"))
}
