// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod fill_holes;
mod value_noise;
mod maze;
mod load_image;
mod vm;
#[cfg(test)]
mod tests;

use std::num::{NonZeroU32, NonZeroUsize};
use std::path::Path;

use serde::Deserialize;
use glam::{UVec2, IVec2, Vec2};
use strum::IntoEnumIterator;

use vm::Vm;

use crate::world::special_area::SpecialArea;
use crate::player::{MAX_HEALTH, ammo_count::AmmoCount};
use crate::powerup::{PowerupType, Effect};

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub(in crate::net) struct Config {
	#[serde(default = "max_players_default")] pub max_players: Option<NonZeroUsize>,
	pub world_size: Vec2,
	#[serde(default)] pub lan_discovery: LanDiscoveryConfig,
	pub blocks: BlockConfig,
	#[serde(default)] pub events: EventConfig,
	#[serde(default)] pub special_areas: SpecialAreaConfig,
}

#[derive(Default, Deserialize)]
#[serde(rename = "LanDiscovery", deny_unknown_fields)]
pub(in crate::net) struct LanDiscoveryConfig {
	#[serde(default)] pub name: String,
	#[serde(default)] pub desc: String,
}

#[derive(Deserialize)]
#[serde(rename = "Blocks", deny_unknown_fields)]
pub(in crate::net) struct BlockConfig {
	pub block_size: NonZeroU32,
	pub grid_size: UVec2,
	pub ops: Box<[Op]>,
}

#[derive(Debug, Deserialize)]
#[serde(deny_unknown_fields)]
pub(in crate::net) enum Op {
	// Binary operations
	Add,
	Sub,
	Mul,
	Div,
	Min,
	Max,

	// Generation
	Constant(f32),
	WhiteNoise { #[serde(default)] seed: u64 },
	ValueNoise {
		#[serde(default)] seed: u64,
		levels: ValueNoiseLevels,
	},
	Maze {
		#[serde(default)] seed: u64,
		#[serde(default = "nonzerou32_one")] wall_thickness: NonZeroU32,
		#[serde(default = "nonzerou32_one")] open_thickness: NonZeroU32,
		#[serde(default)] open_prob: f32,
		#[serde(default)] close_prob: f32,
	},
	Circle(Vec2),
	Rect {
		pos: IVec2,
		size: IVec2,
		#[serde(default = "f32_one")] fill: f32,
	},
	Image(Box<Path>),

	// Utilities
	FillHoles,
	Threshold,

	// Stack operations
	Duplicate(isize),
	Remove(isize),
}

#[derive(Debug, Deserialize)]
#[serde(deny_unknown_fields)]
pub(in crate::net) enum ValueNoiseLevels {
	Fractal {
		min: NonZeroU32,
		scale: NonZeroU32,
		octaves: NonZeroU32,
	},
	Sizes(Box<[NonZeroU32]>),
	WeightedSizes(Box<[(NonZeroU32, f32)]>),
}

#[derive(Deserialize)]
#[serde(rename = "Events", deny_unknown_fields)]
pub struct EventConfig {
	/**
	 * Radius of the square in which the players can spawn, respawn and teleport
	 * to.
	 */
	#[serde(default = "player_spawn_radius_default")] pub player_spawn_radius: f32,
	#[serde(default)] pub ammo_count: AmmoCountIo,
	#[serde(default = "ammo_crate_supply_range_default")] pub ammo_crate_supply_range: (u32, u32),

	/**
	 * There is a fixed probability of an item spawning in a given update,
	 * independent of the previous spawned items, but proportional to the number
	 * of players.
	 *
	 * These variables here are the mean time period for an item to spawn. To
	 * derive the probability of an item spawning in an update, divide delta time
	 * by this and multiply it by the number of players in the game.
	 */
	#[serde(default = "ammo_crate_period_default")] pub ammo_crate_period: f32,
	#[serde(default = "powerup_period_default")] pub powerup_period: f32,

	/**
	 * When an item is spawned, a player is selected at random and the item is
	 * spawned within a square of this radius centred at that player.
	 */
	#[serde(default = "item_spawn_radius_default")] pub item_spawn_radius: f32,

	/**
	 * Specifies a list of effects that powerups are allowed to be.
	 *
	 * When a powerup is spawned, a randomly selected effect is chosen from this
	 * list, which means that if the list contains duplicates, those duplicates
	 * are weighted higher.
	 *
	 * If this list is empty, no powerups are ever spawned.
	 *
	 * By default this is the list of all effects.
	 */
	#[serde(default = "allowed_powerups_default")] pub allowed_powerups: Box<[PowerupType]>,

	#[serde(default = "effect_time_default")] pub effect_time: f32,
	#[serde(default = "speed_mul_default")] pub speed_mul: f32,
	#[serde(default = "other_effect_power_default")] pub reload_mul: f32,
	#[serde(default = "regen_mul_default")] pub regen_mul: f32,
	#[serde(default = "other_effect_power_default")] pub damage_mul: f32,
	#[serde(default = "forcefield_strength_default")] pub forcefield_strength: f32,
	#[serde(default = "health_powerup_health_increase_default")] pub health_powerup_health_increase: f32,
}

impl EventConfig {
	pub fn get_effect_power(&self, effect: Effect) -> f32 {
		match effect {
			Effect::Speed => self.speed_mul,
			Effect::Reload => self.reload_mul,
			Effect::Regeneration => self.regen_mul,
			Effect::Damage => self.damage_mul,
			Effect::Forcefield => self.forcefield_strength,
		}
	}
}

impl Default for EventConfig {
	fn default() -> EventConfig {
		EventConfig {
			player_spawn_radius: player_spawn_radius_default(),
			ammo_count: AmmoCountIo::default(),
			ammo_crate_supply_range: ammo_crate_supply_range_default(),
			ammo_crate_period: ammo_crate_period_default(),
			powerup_period: powerup_period_default(),
			item_spawn_radius: item_spawn_radius_default(),
			allowed_powerups: allowed_powerups_default(),

			effect_time: effect_time_default(),
			speed_mul: speed_mul_default(),
			reload_mul: other_effect_power_default(),
			regen_mul: regen_mul_default(),
			damage_mul: other_effect_power_default(),
			forcefield_strength: forcefield_strength_default(),
			health_powerup_health_increase: health_powerup_health_increase_default(),
		}
	}
}

#[derive(Deserialize)]
#[serde(from = "f64", deny_unknown_fields /* Don't think it's needed */)]
pub struct AmmoCountIo(pub AmmoCount);

impl From<f64> for AmmoCountIo {
	fn from(count: f64) -> AmmoCountIo {
		AmmoCountIo(if count == f64::INFINITY {
			AmmoCount::new_infinite()
		} else {
			AmmoCount::new(count as u32)
		})
	}
}

impl Default for AmmoCountIo {
	fn default() -> AmmoCountIo {
		AmmoCountIo(AmmoCount::new(100))
	}
}

#[derive(Default, Deserialize)]
#[serde(rename = "SpecialAreas", deny_unknown_fields)]
pub struct SpecialAreaConfig {
	#[serde(default)] pub areas: Vec<SpecialArea>,
	#[serde(default)] pub portals: Box<[Portal]>,
	#[serde(default)] pub portal_default_colours: PortalDefaultColourConfig,
	#[serde(default)] pub effect_zones: Box<[EffectZone]>,
	#[serde(default)] pub effect_zone_default_colours: EffectZoneDefaultColourConfig,
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub struct Portal {
	pub p0: Vec2,
	pub p1: Vec2,
	#[serde(default)] pub out: Box<[PortalLink]>,
	pub colour: Option<[u8; 4]>,
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub struct PortalLink {
	pub index: usize,
	#[serde(default)] pub rotation: Rotation,
	#[serde(default)] pub flip_x: bool, // Flipping is done after rotation
	#[serde(default)] pub flip_y: bool,
	#[serde(default = "vec2_one")] pub vel_scl: Vec2,
	#[serde(default)] pub vel_off: Vec2,
}

#[derive(Clone, Copy, Default, PartialEq, Eq, Deserialize)]
#[serde(deny_unknown_fields)]
pub enum Rotation { // Anti-clockwise
	#[default] D0,
	D90,
	D180,
	D270,
}

impl Rotation {
	pub fn rotate(self, v: Vec2) -> Vec2 {
		match self {
			Rotation::D0 => v,
			Rotation::D90 => Vec2::new(-v.y, v.x),
			Rotation::D180 => -v,
			Rotation::D270 => Vec2::new(v.y, -v.x),
		}
	}
}

#[derive(Deserialize)]
#[serde(rename = "PortalDefaultColours", deny_unknown_fields)]
pub struct PortalDefaultColourConfig {
	#[serde(default = "PortalDefaultColourConfig::default_source")] pub source: [u8; 4],
	#[serde(default = "PortalDefaultColourConfig::default_sink")] pub sink: [u8; 4],
	#[serde(default = "PortalDefaultColourConfig::default_bidirectional")] pub bidirectional: [u8; 4],
}

impl PortalDefaultColourConfig {
	fn default_source() -> [u8; 4] { [255, 127, 0, 95] }
	fn default_sink() -> [u8; 4] { [0, 127, 255, 95] }
	fn default_bidirectional() -> [u8; 4] { [127, 0, 255, 95] }
}

impl Default for PortalDefaultColourConfig {
	fn default() -> PortalDefaultColourConfig {
		PortalDefaultColourConfig {
			source: Self::default_source(),
			sink: Self::default_sink(),
			bidirectional: Self::default_bidirectional(),
		}
	}
}

#[derive(Deserialize)]
#[serde(deny_unknown_fields)]
pub struct EffectZone {
	pub p0: Vec2,
	pub p1: Vec2,
	pub effect: Effect,
	pub time: f32,

	/**
	 * Emitting a WorldEvent every single update is unnecessary bandwidth, so
	 * only do it when the remaining time is less than this.
	 */
	pub min_time: f32,
	pub power: Option<f32>, // Falls back to the effect default
	pub colour: Option<[u8; 4]>,
}

#[derive(Deserialize)]
#[serde(rename = "EffectZoneDefaultColours", deny_unknown_fields)]
pub struct EffectZoneDefaultColourConfig {
	#[serde(default = "EffectZoneDefaultColourConfig::default_speed")] pub speed: [u8; 4],
	#[serde(default = "EffectZoneDefaultColourConfig::default_reload")] pub reload: [u8; 4],
	#[serde(default = "EffectZoneDefaultColourConfig::default_regen")] pub regen: [u8; 4],
	#[serde(default = "EffectZoneDefaultColourConfig::default_damage")] pub damage: [u8; 4],
	#[serde(default = "EffectZoneDefaultColourConfig::default_forcefield")] pub forcefield: [u8; 4],
}

impl EffectZoneDefaultColourConfig {
	fn default_speed() -> [u8; 4] { [63, 191, 31, 63] }
	fn default_reload() -> [u8; 4] { [255, 191, 47, 63] }
	fn default_regen() -> [u8; 4] { [255, 127, 223, 63] }
	fn default_damage() -> [u8; 4] { [255, 95, 71, 63] }
	fn default_forcefield() -> [u8; 4] { [63, 191, 255, 63] }
}

impl Default for EffectZoneDefaultColourConfig {
	fn default() -> EffectZoneDefaultColourConfig {
		EffectZoneDefaultColourConfig {
			speed: Self::default_speed(),
			reload: Self::default_reload(),
			regen: Self::default_regen(),
			damage: Self::default_damage(),
			forcefield: Self::default_forcefield(),
		}
	}
}

// It would be nice if serde could inline this in the attribute
fn nonzerou32_one() -> NonZeroU32 { NonZeroU32::new(1).unwrap() }
fn f32_one() -> f32 { 1.0 }
fn vec2_one() -> Vec2 { Vec2::ONE }
fn ammo_crate_period_default() -> f32 { 30.0 }
fn item_spawn_radius_default() -> f32 { 40.0 }
fn powerup_period_default() -> f32 { 60.0 }
fn player_spawn_radius_default() -> f32 { 64.0 }
fn ammo_crate_supply_range_default() -> (u32, u32) { (50, 75) }
fn allowed_powerups_default() -> Box<[PowerupType]> { PowerupType::iter().collect::<Box<_>>() }
fn max_players_default() -> Option<NonZeroUsize> { NonZeroUsize::new(8) }

fn effect_time_default() -> f32 { 30.0 }
fn speed_mul_default() -> f32 { 1.375 }
fn regen_mul_default() -> f32 { 2.0 }
fn forcefield_strength_default() -> f32 { 250.0 }
fn other_effect_power_default() -> f32 { 1.5 }
fn health_powerup_health_increase_default() -> f32 { MAX_HEALTH * 0.5 }

pub(super) type BlockMap = Box<[f32]>;

impl Config {
	pub fn validate(&self) -> Result<(), String> {
		if self.events.ammo_crate_supply_range.0 >= self.events.ammo_crate_supply_range.1 {
			return Err(format!("ammo crate supply range (a, b) requires a ≤ b, got (a, b) = {}, {}", self.events.ammo_crate_supply_range.0, self.events.ammo_crate_supply_range.1));
		}

		Ok(())
	}

	pub fn generate_block_map(&self) -> Result<BlockMap, String> {
		Vm::new(self.blocks.grid_size).generate(&self.blocks.ops)
	}
}
