// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use strum_macros::EnumIter;
use const_format::formatcp;

use super::{CertInfo, endpoint::router::RouterEntry};

pub(super) const INDENTATION: &str = "   ";

pub enum Command {
	Help(Option<CommandVariant>),
	Exit,
	Clear,
	Add(RouterEntry),
	Remove(Box<str>), // Removes by id
	Reload(Box<str>),
	Route(RouteCommand),
	Status(Option<Box<str>>),
	Cert(CertInfo),
	Info,
}

#[derive(Clone, Copy, EnumIter)]
pub enum CommandVariant { Help, Exit, Clear, Add, Remove, Reload, Route, Status, Cert, Info }

impl CommandVariant {
	fn try_from(cmd: &str) -> Option<CommandVariant> {
		match cmd {
			"help" => Some(CommandVariant::Help),
			"exit" => Some(CommandVariant::Exit),
			"clear" => Some(CommandVariant::Clear),
			"add" => Some(CommandVariant::Add),
			"remove" => Some(CommandVariant::Remove),
			"reload" => Some(CommandVariant::Reload),
			"route" => Some(CommandVariant::Route),
			"status" => Some(CommandVariant::Status),
			"cert" => Some(CommandVariant::Cert),
			"info" => Some(CommandVariant::Info),
			_ => None,
		}
	}

	pub fn help(self) -> &'static str {
		use self::INDENTATION as I;

		match self {
			CommandVariant::Help => formatcp!("help - prints help text\n{I}help: Prints the help text of all comamnds.\n{I}help <COMMAND>: Prints the help text of a particular command.\n"),
			CommandVariant::Exit => "exit - exits the server\n",
			CommandVariant::Clear => "clear - clears the screen (you can also do this with Ctrl+L)\n",
			CommandVariant::Add => formatcp!("add - adds a new game server\n{I}add <ID>:<ID>:...:<ID>:<PATH>: Loads a game server from PATH with ID as each id. If there are no ids, the id \"default\" is used, which routes everything else to it.\n"),
			CommandVariant::Remove => formatcp!("remove - removes an existing game server\n{I}remove <ID>: Removes the game server with id of ID.\n"),
			CommandVariant::Reload => formatcp!("reload - reloads the configuration of an existing game server\n{I}reload <ID>: Creates a new game server using the configuration file from the server with ID. Then replaces that server with the newly loaded server on success, preserving all routing information.\n"),
			CommandVariant::Route => formatcp!("route - change routing information\n{I}route add <ID> <NEW_ID>: Routes NEW_ID to the game server with id of ID. ID must correspond with an existing game server and NEW_ID must not already be used.\n{I}route remove <ID>: Removes the routing rule from ID to its corresponding game server. ID must exist and the server with that id must have another id so it's still reachable.\n"),
			CommandVariant::Status => formatcp!("status - prints information about running game servers\n{I}status: Prints information about all servers.\n{I}status <ID>: Prints information about a server with the id of ID.\n"),
			CommandVariant::Cert => formatcp!("cert - updates certificate information\n{I}cert <CERT_PATH> <KEY_PATH>: Updates the TLS certificate used by the server, taking paths to files in the PEM format.\n"),
			CommandVariant::Info => "info - prints the publicly available build information about the server\n",
		}
	}
}

pub enum RouteCommand {
	Add {
		id: Box<str>,
		new_id: Box<str>,
	},
	Remove(Box<str>),
}

impl Command {
	pub fn build(line: &str) -> Result<Option<Command>, String> {
		let tokens = shell_words::split(line).map_err(|err| format!("failed parsing line: {err}"))?;
		let (cmd, args) = match tokens.split_at_checked(1) {
			Some((first, last)) => (&first[0], last),
			None => return Ok(None),
		};

		CommandBuilder { cmd, args }.build()
	}
}

struct CommandBuilder<'a> {
	cmd: &'a String,
	args: &'a [String],
}

impl CommandBuilder<'_> {
	fn build(self) -> Result<Option<Command>, String> {
		Ok(Some(match CommandVariant::try_from(self.cmd.as_str()) {
			Some(CommandVariant::Help) => {
				self.check_at_most_n_args(1)?;

				let variant = if let Some(cmd) = self.args.first() {
					if let Some(variant) = CommandVariant::try_from(cmd) {
						Some(variant)
					} else {
						return Err(format!("help: unknown command of \"{cmd}\""));
					}
				} else { None };

				Command::Help(variant)
			},
			Some(CommandVariant::Exit) => self.require_no_args(Command::Exit)?,
			Some(CommandVariant::Clear) => self.require_no_args(Command::Clear)?,
			Some(CommandVariant::Add) => {
				let game = &self.require_n_args::<1>()?[0];
				Command::Add(RouterEntry::build(game).map_err(|err| format!("add: {err}"))?)
			},
			Some(CommandVariant::Remove) => Command::Remove(self.require_n_args::<1>()?[0].clone().into_boxed_str()),
			Some(CommandVariant::Reload) => Command::Reload(self.require_n_args::<1>()?[0].clone().into_boxed_str()),
			Some(CommandVariant::Route) => {
				let subcmd = self.get_arg(0)?;
				match subcmd {
					"add" => {
						let [_, id, new_id] = self.require_n_args_subcmd::<3>(subcmd)?;
						Command::Route(RouteCommand::Add { id: id.clone().into_boxed_str(), new_id: new_id.clone().into_boxed_str() })
					},
					"remove" => {
						let [_, id] = self.require_n_args_subcmd::<2>(subcmd)?;
						Command::Route(RouteCommand::Remove(id.clone().into_boxed_str()))
					},
					_ => return Err(format!("route: unknown subcommand of \"{subcmd}\"")),
				}
			},
			Some(CommandVariant::Status) => {
				self.check_at_most_n_args(1)?;
				Command::Status(self.args.first().map(|id| id.clone().into_boxed_str()))
			},
			Some(CommandVariant::Cert) => {
				let [cert_path, key_path] = self.require_n_args::<2>()?;
				Command::Cert(CertInfo { cert_path: cert_path.clone().into_boxed_str(), key_path: key_path.clone().into_boxed_str() })
			},
			Some(CommandVariant::Info) => self.require_no_args(Command::Info)?,
			None => return Err(format!("unknown command of \"{}\", run `help` for more info", self.cmd)),
		}))
	}

	fn require_no_args(&self, command: Command) -> Result<Command, String> {
		if self.args.is_empty() {
			Ok(command)
		} else {
			Err(format!("{}: expected no arguments", self.cmd))
		}
	}

	fn require_n_args<const N: usize>(&self) -> Result<&[String; N], String> {
		self.args.try_into().map_err(|_| format!("{}: expected {N} argument(s), got {}", self.cmd, self.args.len()))
	}

	fn require_n_args_subcmd<const N: usize>(&self, subcmd: &str) -> Result<&[String; N], String> {
		self.args.try_into().map_err(|_| format!("{}: {subcmd}: expected {N} argument(s), got {}", self.cmd, self.args.len()))
	}

	fn check_at_most_n_args(&self, count: usize) -> Result<(), String> {
		if self.args.len() <= count {
			Ok(())
		} else {
			Err(format!("{}: expected at most {count} argument(s), got {}", self.cmd, self.args.len()))
		}
	}

	fn get_arg(&self, index: usize) -> Result<&str, String> {
		self.args.get(index).map_or_else(|| Err(format!("{}: expected at least {} argument(s), got {}", self.cmd, index + 1, self.args.len())), |arg| Ok(arg.as_str()))
	}
}

#[cfg(test)]
mod tests {
	use strum::IntoEnumIterator;

	use super::*;

	#[test]
	fn all_help_text_ends_in_newline() {
		for cmd in CommandVariant::iter() {
			assert!(cmd.help().ends_with('\n'));
		}
	}
}
