// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use egui::{Color32, TextEdit};
use glium::Frame as GliumFrame;

use super::common::{Common, CommonGs};

use crate::game::Game;
use crate::gui::{GuiState, style::{UiExt, FontSize, H4_SIZE}};
use crate::net::{Address, server::PORT};

pub struct PsEnterAddress {
	common: Common,
	host: String,
	port: String,
	game_id: String,
	prev_was_u16: bool,
}

impl PsEnterAddress {
	pub fn new() -> PsEnterAddress {
		PsEnterAddress {
			common: Common::new(),
			host: String::from("127.0.0.1"),
			port: PORT.to_string(),
			game_id: String::new(),
			prev_was_u16: true,
		}
	}
}

impl CommonGs for PsEnterAddress {
	fn common(&mut self) -> &mut Common { &mut self.common }

	fn loop_iter(&mut self, game: &mut Game, frame: &mut GliumFrame) -> bool {
		GuiState::update(&mut game.gui, "ps-enter-address", &game.window, |ctx, ui| {
			ui.h1("Enter Address").no_bottom_padding().add();

			ui.h3("Host").add();
			ui.add(TextEdit::singleline(&mut self.host).font_size(H4_SIZE));

			ui.h3("Game ID").add();
			ui.add(TextEdit::singleline(&mut self.game_id).font_size(H4_SIZE));

			ui.h3("Port").add();

			let style = ui.style_mut();
			let orig_bg = style.visuals.extreme_bg_color;
			if !self.prev_was_u16 {
				style.visuals.extreme_bg_color = Color32::from_rgb(0x5f, 0x00, 0x00);
			}
			ui.add(TextEdit::singleline(&mut self.port).font_size(H4_SIZE));
			ui.style_mut().visuals.extreme_bg_color = orig_bg;
			ui.add_space(16.0);

			let port = self.port.parse().ok();
			self.prev_was_u16 = port.is_some();
			if let (Some(starter), Some(port)) = (self.common.connection_starter(), port) {
				if ui.b2("Play").clicked() {
					let addr = Address { host: Box::from(self.host.as_str()), port, game_id: Box::from(self.game_id.as_str()) };
					starter.connect(game.endpoint.get(&game.tokio_runtime), &game.tokio_runtime, addr);
				}
			} else {
				ui.b2_enabled("Play", false);
			}

			self.common.finish_update(ctx, ui);
		});

		game.gui.render(&game.display, frame);

		false
	}
}
