// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use serde::{Serialize, Deserialize};
use strum::{EnumCount, IntoEnumIterator};

use super::Effect;

use crate::world::DMap;
#[cfg(feature = "client")] use crate::world::MAX_PLAYERS;

#[derive(Debug, Default, Clone, Copy, Serialize, Deserialize)]
pub struct EffectInfo {
	pub time: f32,
	pub power: f32,
}

impl EffectInfo {
	fn active(self) -> bool {
		self.time > 0.0
	}

	fn get_power(self) -> Option<f32> {
		self.active().then_some(self.power)
	}
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct ActiveEffects(DMap<u32, [EffectInfo; Effect::COUNT]>);

impl ActiveEffects {
	pub fn new() -> ActiveEffects {
		ActiveEffects(DMap::new())
	}

	#[cfg(feature = "client")]
	pub fn validate(&self) -> Result<(), &'static str> {
		if self.0.len() <= MAX_PLAYERS { Ok(()) }
		else { Err("too many players") }
	}

	pub fn remove_player(&mut self, id: u32) {
		self.0.remove(&id);
	}

	pub fn get(&self, id: u32, effect: Effect) -> Option<EffectInfo> {
		self.0.get(&id).map(|effects| effects[effect as usize])
	}

	/**
	 * Returns the "power" of the effect if the player has that effect, otherwise
	 * None.
	 *
	 * The interpretation of this "power" depends on the type of effect. For
	 * forcefields, it's the strength of the repulsion and for everything else
	 * it's a multiplier.
	 */
	pub fn get_power(&self, id: u32, effect: Effect) -> Option<f32> {
		self.0.get(&id)?[effect as usize].get_power()
	}

	/**
	 * Returns the remaining time of the effect, or zero if the effect isn't
	 * applied.
	 */
	pub fn get_time(&self, id: u32, effect: Effect) -> f32 {
		self.0.get(&id).map_or(0.0, |effects| effects[effect as usize].time)
	}

	/**
	 * Convenience method for effects when the power is a multiplier, to which a
	 * default of 1 is returned if the effect isn't active.
	 */
	pub fn get_mul(&self, id: u32, effect: Effect) -> f32 {
		self.get_power(id, effect).unwrap_or(1.0)
	}

	pub fn add_effect(&mut self, id: u32, effect: Effect, time: f32, power: f32) {
		let effects = self.0.entry(id).or_default();
		effects[effect as usize] = EffectInfo { time, power };
	}

	pub fn active_iter(&self, id: u32) -> Option<impl Iterator<Item = Effect> + use<'_>> {
		let effects = self.0.get(&id)?;
		Some(Effect::iter().filter(|&effect| effects[effect as usize].active()))
	}

	/**
	 * Returns an iterator of all player ids and the corresponding time of all
	 * players with `effect`.
	 */
	pub fn player_iter_with_effect(&self, effect: Effect) -> impl Iterator<Item = (u32, EffectInfo)> + '_ {
		self.0.iter().filter_map(move |(&id, effects)| {
			let info = effects[effect as usize];
			info.active().then_some((id, info))
		})
	}

	pub fn update(&mut self, dt: f32) {
		self.0.retain(|_, effects| {
			let mut has_effects = false;
			for effect in effects {
				effect.time -= dt;
				has_effects |= effect.active();
			}
			has_effects
		});
	}
}
