// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
#[cfg(feature = "client")] pub mod renderer;

use std::collections::BTreeMap;

use serde::{Serialize, Deserialize};

use crate::net::message::{Message, ScoreMessage};
#[cfg(feature = "client")] use crate::{world::MAX_PLAYERS, net::client::ClientError};

#[derive(Default, Clone, Serialize, Deserialize)]
pub struct Scores(BTreeMap<u32, i32>);

impl Scores {
	#[cfg(feature = "client")]
	pub fn validate(&self) -> Result<(), ClientError> {
		if self.0.len() <= MAX_PLAYERS {
			Ok(())
		} else {
			Err(ClientError::Limit(String::from("too many scores")))
		}
	}

	#[cfg(feature = "client")]
	pub fn set(&mut self, id: u32, score: Option<i32>) {
		if let Some(score) = score {
			self.0.insert(id, score);
		} else {
			self.0.remove(&id);
		}
	}

	/**
	 * Adds a new player with a score of 0 if that player doesn't exist, or
	 * increments their score if they do exist.
	 */
	pub fn add(&mut self, id: u32) -> Message {
		let score = self.0.entry(id).and_modify(|score| *score += 1).or_insert(0);
		Message::Score(ScoreMessage::Score(id, Some(*score)))
	}

	pub fn remove(&mut self, id: u32) -> Message {
		self.0.remove(&id);
		Message::Score(ScoreMessage::Score(id, None))
	}
}
