// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use std::sync::Arc;

use egui::{Id, SidePanel, Frame, Ui, Margin, FontId, FontFamily, TextStyle, RichText, TextFormat, Color32, Context, text::LayoutJob};
use glam::Vec3;

use super::Scores;

use crate::player::PlayerRenderer;

const GREY: u8 = 0xdf;

pub fn render(ctx: &Context, scores: &Scores, player_id: u32) {
	SidePanel::left(Id::new("panel"))
		.frame(Frame::none().inner_margin(Margin::same(16.0)))
		.exact_width(f32::MAX)
		.show_separator_line(false)
		.show(ctx, |ui| {
			ui.label(RichText::new("Scores").color(Color32::from_gray(GREY)).text_style(TextStyle::Name(Arc::from("scores"))));
			ui.add_space(8.0);

			let mut rendered_scores = scores.0.iter().map(|(&id, &score)| (id, score)).collect::<Vec<_>>();
			rendered_scores.sort_by_key(|e| !e.1); // Sorts by scores in descending order (not rather than negation to avoid overflow with i32::MIN), if equal the BTreeMap results in them being then sorted by id in ascending order
			rendered_scores.truncate(16);

			let rs_has_current_player = rendered_scores.iter().any(|e| e.0 == player_id);
			let truncated = rendered_scores.len() < scores.0.len();

			for (id, score) in rendered_scores {
				add_text(id, score, player_id, ui);
				ui.add_space(4.0);
			}

			if truncated { // If the scores are truncated
				ui.colored_label(Color32::from_gray(GREY), "…");
			}

			// Always shows the current player's score even if not in the top
			if let Some(&score) = scores.0.get(&player_id) {
				if !rs_has_current_player {
					ui.add_space(4.0);
					add_text(player_id, score, player_id, ui);
				}
			}
		});
}

fn add_text(id: u32, score: i32, player_id: u32, ui: &mut Ui) {
	let colour = Vec3::from_array(PlayerRenderer::get_colour(id)).max(Vec3::splat(0.375));
	let mut job = LayoutJob::default();
	job.append(&score.to_string(), if id == player_id { 8.0 } else { 0.0 }, TextFormat { // Indents the score of the current player
		font_id: FontId::new(14.0, FontFamily::Name("scores".into())),
		color: Color32::from_rgb((colour.x * 255.0) as u8, (colour.y * 255.0) as u8, (colour.z * 255.0) as u8),
		.. TextFormat::default()
	});
	ui.label(job);
}
