// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use crate::camera::Camera;

use glam::{Vec2, Vec3, swizzles::Vec3Swizzles};

const FADE_IN_TIME: f32 = 4.0;
const Z_MIN: f32 = 1.0; // Multiplied by zoom
const Z_LEN: f32 = 400.0;
const Z_FADE_DIST: f32 = 4.0;

#[derive(Debug)]
pub struct Star {
	pos: Vec3,
	vel: Vec3,
	int: f32,
	time: f32,
}

impl Star {
	pub fn new(camera: &Camera) -> Star {
		Star::reset(camera, f32::INFINITY, None)
	}

	fn reset(camera: &Camera, init_time: f32, z: Option<f32>) -> Star {
		let size = camera.get_size();
		let pos = Vec2::new(rand::random::<f32>() - 0.5, rand::random::<f32>() - 0.5) * size;
		let zoom = size.min_element();
		let z = z.unwrap_or_else(|| rand::random::<f32>() * Z_LEN + Z_MIN * zoom);

		Star {
			pos: (pos * z / zoom).extend(z),
			vel: Vec3::new(rand::random::<f32>() - 0.5, rand::random::<f32>() - 0.5, rand::random::<f32>() - 0.5) * 0.75,
			int: rand::random::<f32>() * 0.375 + 0.625,
			time: init_time,
		}
	}

	pub fn get_rendered_pos(&self, zoom: f32) -> Vec2 {
		self.pos.xy() / self.pos.z * zoom
	}

	pub fn get_alpha(&self, zoom: f32) -> f32 {
		let mut alpha = self.int // Intrinsic intensity of a star
		/ (self.pos.z.powi(2) * 5e-5 + 1.0) // Attenuation due to distance
		* (self.time / FADE_IN_TIME).min(1.0); // Fade in for new star

		// Fades out a star when its z coordinate is about to move out of bounds, to avoid it suddenly disappearing
		if self.pos.z < Z_MIN * zoom + Z_FADE_DIST {
			alpha *= (-Z_MIN * zoom + self.pos.z) / Z_FADE_DIST;
		} else if self.pos.z > Z_MIN * zoom + Z_LEN - Z_FADE_DIST {
			alpha *= (Z_MIN * zoom + Z_LEN - self.pos.z) / Z_FADE_DIST;
		}

		alpha
	}

	pub fn resize(&mut self, old_size: Vec2, new_size: Vec2) {
		let old_zoom = old_size.min_element();
		let new_zoom = new_size.min_element();

		let r_pos = self.get_rendered_pos(old_zoom);
		self.pos.z += Z_MIN * (new_zoom - old_zoom);
		for i in 0..2 {
			self.pos[i] = r_pos[i] * self.pos.z / new_zoom * new_size[i] / old_size[i];
		}
	}

	pub fn move_by(&mut self, off: Vec2) {
		self.pos.x += off.x;
		self.pos.y += off.y;
	}

	pub fn update(&mut self, dt: f32) {
		self.pos += self.vel * dt;
		self.time += dt;
	}

	/**
	 * Checks if the position is out of bounds, in which it is wrapped around
	 * (returning true), or reset (returning false).
	 *
	 * If the z coordinate is out of bounds, the star is reset.
	 *
	 * If the x or y coordinate is out of bounds, the star is attempted to be
	 * wrapped around, but it might be reset.
	 *
	 * If the star being wrapped around will result in it being "stuck" at the
	 * edges/corner, then it is reset. A situation in which this could happen
	 * for example is if the velocity is something like (0, 0, -z) in which the
	 * star hitting the edge is due to it going out of the camera's view as it
	 * moves closer towards it. If the star was wrapped around in this case, it
	 * wouldn't move around like normal but instead move in the opposite
	 * direction (away from the centre again) resulting in another wrap.
	 *
	 * To detect this, I am checking if the sign star's "effective velocity",
	 * v_i for i = x or y for the components, has the same sign as the new
	 * position. If it does (or there's a zero), the star will be reset.
	 */
	pub fn out_of_bounds(&mut self, camera: &Camera, moved_vel: Vec2) {
		let size = camera.get_size();
		let zoom = size.min_element();

		if self.pos.z < zoom * Z_MIN || self.pos.z > zoom * Z_MIN + Z_LEN {
			*self = Star::reset(camera, 0.0, None);
			return;
		}

		let mut r_pos = self.get_rendered_pos(zoom);
		for i in 0..2 {
			if r_pos[i].abs() > size[i] / 2.0 {
				let x = r_pos[i] / size[i] + 0.5;
				r_pos[i] = (x - x.floor() - 0.5) * size[i];
				self.pos[i] = r_pos[i] * self.pos.z / zoom;

				/*
				 * Effective velocity is calculated as follows:
				 *
				 * ev = d/dt ep
				 *    = d/dt (p_i / p_z) (ignoring zoom but that's a constant)
				 *    = (v_i * p_z - p_i * v_z) / p_z² (quotient rule)
				 *
				 * Note that the denominator is always non-negative so it can be
				 * removed when calculating the sign.
				 */
				let eff_vel = (self.vel[i] + moved_vel[i]) * self.pos.z - self.pos[i] * self.vel.z; // A value with the same sign as the effective velocity
				if eff_vel * self.pos[i] >= 0.0 {
					*self = Star::reset(camera, 0.0, Some(self.pos.z));
					return;
				}
			}
		}
	}
}
