// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
pub mod changes;
#[cfg(feature = "client")] mod border;
#[cfg(feature = "client")] pub mod smooth;
#[cfg(feature = "client")] mod renderer;
pub mod update;
pub mod event;
pub mod special_area;

#[cfg(feature = "client")] pub use smooth::SmoothWorld;
#[cfg(feature = "client")] pub use renderer::Renderer as WorldRenderer;
use special_area::SpecialArea;

use std::collections::{BTreeMap, BTreeSet};

use serde::{Serialize, Deserialize};
use glam::Vec2;
use rand::SeedableRng;
use rand_xoshiro::Xoshiro128StarStar;

use crate::player::{Player, forcefield::Forcefield};
use crate::bullet::BulletManager;
use crate::ammo_crate::AmmoCrate;
use crate::powerup::{Powerup, Effect, active_effects::ActiveEffects};
use crate::utils::{maths::Circle, smallest_unused};

/*
 * Types for deterministic maps and sets.
 *
 * One key idea in this game is that after connecting, the server only needs to
 * send the relative state of the game to the client, and that the client can
 * reproduce an identical state to that of the server.
 *
 * If not, then the client and server Worlds would become out of sync, which
 * depending how out of sync things get, the client could get a completely
 * inaccurate reflection on how the game is going.
 *
 * To satisfy this requirement of determinism, I must be very careful about
 * using HashMaps and HashSets because of their arbitrary iteration order being
 * dependent on some internal random state (for security against DoS attacks).
 *
 * I could use HashMaps and HashSets but remember not to iterate over them in
 * cases when the iteration order is important, but that is trying to write
 * deterministic code in a non-deterministic system. That is, similar to
 * programming in C being trying to write safe code in an unsafe system, which
 * is why Rust exists.
 *
 * This is why I'm instead using BTreeMaps and BTreeSets in all parts of the
 * code that are executed by both the client and server as they have guarantees
 * over iterating through items in sorted order of keys (the specific order
 * isn't important, just the fact that an agreed upon order exists).
 *
 * I am using the type aliases DMap and DSet rather than indirectly using
 * BTreeMap and BTreeSet because in the future I might decide to change the
 * implementation, and doing it here means that all code using deterministic
 * maps and sets will be updated. If I did do this, this would be an update to
 * the protocol with client/server nodes of differing versions possibly becoming
 * out of sync.
 *
 * This requirement of determinism is why I'm using a fixed delta time for
 * physics updates rather than using the actual time between frames.
 */
pub type DMap<K, V> = BTreeMap<K, V>;
pub type DSet<V> = BTreeSet<V>;

pub type WorldRng = Xoshiro128StarStar;

const UPDATES_PER_SECOND: f32 = 240.0;
pub const DELTA_TIME: f32 = 1.0 / UPDATES_PER_SECOND;

// At most 60 seconds to avoid DoS attacks clogging up queues
pub const MAX_STEPS: usize = (60.0 * UPDATES_PER_SECOND) as usize;

// Arbitrary limits used by the client to avoid application layer denial of service attacks
#[cfg(feature = "client")] pub const MAX_PLAYERS: usize = 1024;
#[cfg(feature = "client")] pub const MAX_BULLETS: usize = 16384;
#[cfg(feature = "client")] pub const MAX_SPECIAL_AREAS: usize = 4096;
const MAX_AMMO_CRATES: usize = 512;
const MAX_POWERUPS: usize = 512;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct World {
	size: Vec2,
	players: DMap<u32, Player>,
	bullets: BulletManager,
	ammo_crates: Vec<AmmoCrate>,
	powerups: Vec<Powerup>,
	effects: ActiveEffects,
	rng: WorldRng,
	special_areas: Box<[SpecialArea]>,
}

impl World {
	pub fn new(size: Vec2, special_areas: Box<[SpecialArea]>) -> World {
		World {
			size,
			players: DMap::new(),
			bullets: BulletManager::new(),
			ammo_crates: Vec::new(),
			powerups: Vec::new(),
			effects: ActiveEffects::new(),
			rng: WorldRng::seed_from_u64(rand::random::<u64>()),
			special_areas,
		}
	}

	fn new_player(&mut self) {
		self.players.insert(smallest_unused::find(&self.players), Player::new());
	}

	/** Validation that just needs to be performed once. */
	#[cfg(feature = "client")]
	pub fn validate_init(&self) -> Result<(), &'static str> {
		if self.special_areas.len() > MAX_SPECIAL_AREAS {
			return Err("too many special areas in world");
		}

		Ok(())
	}

	/** Validation that needs to be performed every update. */
	#[cfg(feature = "client")]
	pub fn validate(&self) -> Result<(), &'static str> {
		if !self.size.cmpge(Vec2::splat(0.0)).all() { return Err("invalid world size"); } // Need to do ! >= rather than <, as the latter allows a world size of NaN
		if self.players.len() > MAX_PLAYERS { return Err("too many players in world"); }
		if self.bullets.len() > MAX_BULLETS { return Err("too many bullets in world"); }
		if self.ammo_crates.len() > MAX_AMMO_CRATES { return Err("too many ammo crates in world"); }
		if self.powerups.len() > MAX_POWERUPS { return Err("too many powerups in world"); }
		self.effects.validate()
	}

	pub fn player_count(&self) -> usize { self.players.len() }
	#[cfg(feature = "client")] pub fn get_ammo_string(&self, id: u32) -> String { self.players[&id].get_ammo_count().to_string() }
	#[cfg(feature = "client")] fn get_ammo_crates(&self) -> &[AmmoCrate] { &self.ammo_crates }
	#[cfg(feature = "client")] pub fn get_powerups(&self) -> &[Powerup] { &self.powerups }

	fn get_forcefields(&self) -> Vec<Forcefield> {
		self.effects
			.player_iter_with_effect(Effect::Forcefield)
			.map(|(id, info)| Forcefield { id, pos: self.players[&id].get_pos(), strength: info.power })
			.collect()
	}

	#[cfg(feature = "client")]
	pub fn has_player(&self, id: u32) -> bool { self.players.contains_key(&id) }
}
