// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
#[cfg(feature = "client")] use glam::Vec2;

use super::DSet;

use crate::bullet::Bullet;
use crate::player::Player;
use crate::powerup::{PowerupType, Effect};
#[cfg(feature = "client")] use crate::player::direction::Direction;
#[cfg(feature = "client")] use crate::utils::maths::Circle;

/*
 * Both the client and server need to know certain things that have changed
 * after updating the world. However, the exact information that's relevant to
 * them differs.
 *
 * As an optimisation, generics are used to avoid performing unnecessary
 * operations when calling `update`.
 */
pub trait Changes {
	fn add_hit(&mut self, bullet: Bullet, player: &Player, player_id: u32, killed: bool);
	fn pos_changed(&mut self, id: u32);
	fn ammo_crate_collected(&mut self, id: u32);
	fn powerup_collected(&mut self, id: u32, typ: PowerupType);
}

#[cfg(feature = "client")]
pub struct Hit {
	pub bullet: Bullet,
	pub player_id: u32,
	pub player_pos: Vec2,
	pub player_vel: Vec2,
	pub player_dir: Direction,
	pub killed: bool,
}

#[cfg(feature = "client")]
#[derive(Default)]
pub struct ClientChanges { // Used by the clients
	pub moved: DSet<u32>, // Whenever the server moves a player, which can be on death or teleportation
	pub hits: Vec<Hit>,
}

#[cfg(feature = "client")]
impl Changes for ClientChanges {
	fn add_hit(&mut self, bullet: Bullet, player: &Player, player_id: u32, killed: bool) {
		self.hits.push(Hit {
			bullet,
			player_id,
			player_pos: player.get_pos(),
			player_vel: player.get_vel(),
			player_dir: player.get_dir(),
			killed,
		});
	}

	fn pos_changed(&mut self, id: u32) { self.moved.insert(id); }
	fn ammo_crate_collected(&mut self, _id: u32) {}
	fn powerup_collected(&mut self, _id: u32, _typ: PowerupType) {}
}

#[derive(Default)]
pub struct ServerChanges {
	pub killers: Vec<u32>, // Player ids duplicated if they get multiple kills in a single update
	pub(super) randomly_moved: DSet<u32>,
	pub(super) dead: DSet<u32>,
	pub(super) ammo_crates_collected: Vec<u32>,
	pub(super) effects_to_apply: Vec<(u32, Effect)>,
}

impl Changes for ServerChanges {
	fn add_hit(&mut self, bullet: Bullet, _player: &Player, player_id: u32, killed: bool) {
		if killed {
			if let Some(owner) = bullet.get_owner() {
				self.killers.push(owner);
			}

			self.randomly_moved.insert(player_id);
			self.dead.insert(player_id);
		}
	}

	fn pos_changed(&mut self, _id: u32) {}
	fn ammo_crate_collected(&mut self, id: u32) { self.ammo_crates_collected.push(id); }

	fn powerup_collected(&mut self, id: u32, typ: PowerupType) {
		if let Some(effect) = Effect::try_from_type(typ) {
			self.effects_to_apply.push((id, effect));
		} else if typ == PowerupType::Teleportation {
			self.randomly_moved.insert(id);
		}
	}
}

pub struct IgnoreChanges;

impl Changes for IgnoreChanges {
	fn add_hit(&mut self, _bullet: Bullet, _player: &Player, _player_id: u32, _killed: bool) {}
	fn pos_changed(&mut self, _id: u32) {}
	fn ammo_crate_collected(&mut self, _id: u32) {}
	fn powerup_collected(&mut self, _id: u32, _typ: PowerupType) {}
}
