// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod move_chooser;
pub mod portal;
pub mod effect_zone;

use std::collections::BTreeSet;

use serde::{Serialize, Deserialize};
use glam::Vec2;
use rand::{SeedableRng, RngCore};

use move_chooser::MoveChooser;
use portal::PortalManager;
use effect_zone::EffectZones;

use super::{World, WorldRng, changes::{Changes, ServerChanges, IgnoreChanges}, DELTA_TIME, MAX_AMMO_CRATES, MAX_POWERUPS};

use crate::player::{Player, ammo_count::AmmoCount};
use crate::blocks::Blocks;
use crate::ammo_crate::{AmmoCrate, RADIUS as AMMO_CRATE_RADIUS};
use crate::net::server::config::EventConfig;
use crate::powerup::{Powerup, PowerupType, Effect, RADIUS as POWERUP_RADIUS};
use crate::utils::{rng::Float, maths::{self, Circle, CollidingRect, RectCorners}};

#[derive(Clone, Serialize, Deserialize)]
pub enum WorldEvent {
	/// When a new player joins.
	NewPlayer, // No id included as that can be inferred from the current players (smallest unused)

	/// Sets the player's position.
	PlayerPos(u32, Vec2),

	/// Sets the player's velocity.
	PlayerVel(u32, Vec2),

	/// Sets the player's amount of ammo.
	PlayerAmmo(u32, AmmoCount),

	/// Sets the player's health.
	PlayerHealth(u32, f32),

	/// Adds an effect to a player.
	PlayerEffect {
		id: u32,
		effect: Effect,
		time: f32,
		power: f32,
	},

	/// Adds a new ammo crate. The amount of ammo is determined when the player
	/// picks it up, in which the server may give the player ammo with the
	/// `PlayerAmmo` update.
	NewAmmoCrate(Vec2),

	/// Adds a new powerup.
	NewPowerup(Vec2, PowerupType),
}

pub struct EventChooser {
	rng: WorldRng,
	move_chooser: MoveChooser,
	portal_manager: PortalManager,
	effect_zones: EffectZones,
	config: EventConfig,
}

impl EventChooser {
	pub fn build(world: &World, blocks: &Blocks, portal_manager: PortalManager, effect_zones: EffectZones, config: EventConfig) -> Result<EventChooser, String> {
		let mut rng = WorldRng::seed_from_u64(rand::random::<u64>());
		let move_chooser = MoveChooser::build(world, blocks, config.player_spawn_radius, &mut rng)?;
		Ok(EventChooser { rng, move_chooser, portal_manager, effect_zones, config })
	}

	pub fn get_pre_events(&mut self, world: &World, blocks: &Blocks, new_players: Option<&BTreeSet<u32>>) -> Vec<WorldEvent> {
		/*
		 * Also using the client limits to prevent clients from disconnecting if
		 * there are too many. Not also doing this with bullets as that case is
		 * less likely to occur (this can occur by players staying idle for a
		 * while) and players not being able to shoot would be weird.
		 */
		let mut events = Vec::new();

		if self.rng.next_f32() < EventChooser::get_prob(self.config.ammo_crate_period, world.players.len()) && world.ammo_crates.len() < MAX_AMMO_CRATES {
			let can_spawn = |pos| {
				let rect = RectCorners(pos - AMMO_CRATE_RADIUS, pos + AMMO_CRATE_RADIUS);
				!(blocks.iter_range(pos, AMMO_CRATE_RADIUS).count() > 0 || self.portal_manager.sink_colliding(&rect))
			};
			if let Some(pos) = EventChooser::find_item_pos(world, &mut self.rng, &self.config, AMMO_CRATE_RADIUS, can_spawn) {
				events.push(WorldEvent::NewAmmoCrate(pos));
			}
		}

		if self.rng.next_f32() < EventChooser::get_prob(self.config.powerup_period, world.players.len()) && world.powerups.len() < MAX_POWERUPS && !self.config.allowed_powerups.is_empty() {
			let can_spawn = |pos| {
				let circle = (pos, POWERUP_RADIUS);
				!(blocks.iter_range(pos, POWERUP_RADIUS).any(|block| circle.colliding_rect(block.get_rect())) || self.portal_manager.sink_colliding(&circle))
			};
			if let Some(pos) = EventChooser::find_item_pos(world, &mut self.rng, &self.config, POWERUP_RADIUS, can_spawn) {
				let effect = self.config.allowed_powerups[self.rng.next_u32() as usize % self.config.allowed_powerups.len()];
				events.push(WorldEvent::NewPowerup(pos, effect));
			}
		}

		if let Some(new_players) = new_players {
			for &id in new_players {
				events.push(WorldEvent::NewPlayer);
				events.push(WorldEvent::PlayerPos(id, self.move_chooser.generate(blocks, &mut self.rng)));
				events.push(WorldEvent::PlayerAmmo(id, self.config.ammo_count.0));
			}
		}

		events
	}

	pub fn get_and_apply_post_events(&mut self, world: &mut World, blocks: &Blocks, changes: &ServerChanges) -> Vec<WorldEvent> {
		let mut events = Vec::new();

		// Immediately applies the event then pushes it
		// Using a macro to avoid borrow checker problems
		macro_rules! push {
			( $x:expr ) => {
				{
					let event = $x;
					world.apply_event(&event, &mut IgnoreChanges);
					events.push(event);
				}
			};
		}

		for &id in &changes.randomly_moved {
			push!(WorldEvent::PlayerPos(id, self.move_chooser.generate(blocks, &mut self.rng)));
			self.portal_manager.on_random_move(id);
		}

		for &id in &changes.dead {
			push!(WorldEvent::PlayerVel(id, Vec2::ZERO));
			push!(WorldEvent::PlayerAmmo(id, self.config.ammo_count.0));

			// Resets all effects when a player dies
			if let Some(effects) = world.effects.active_iter(id).map(Iterator::collect::<Vec<_>>) {
				for effect in effects {
					push!(WorldEvent::PlayerEffect { id, effect, time: 0.0, power: 0.0 });
				}
			}
		}

		{
			let mut tmp_events = Vec::new();
			self.portal_manager.events(world, blocks, &mut self.rng, &mut tmp_events);
			self.effect_zones.events(world, &mut tmp_events);
			for event in tmp_events { push!(event); }
		}

		for &id in &changes.ammo_crates_collected {
			if changes.dead.contains(&id) { continue; }

			let range = self.config.ammo_crate_supply_range;
			let (min, max) = (range.0, range.1);
			let new_ammo = world.players[&id].get_ammo_count() + (min + self.rng.next_u32() % ((max - min).saturating_add(1)));
			push!(WorldEvent::PlayerAmmo(id, new_ammo));
		}

		for &(id, effect) in &changes.effects_to_apply {
			if changes.dead.contains(&id) { continue; }

			let power = self.config.get_effect_power(effect);

			if effect == Effect::Regeneration {
				push!(WorldEvent::PlayerHealth(id, world.players[&id].add_health(self.config.health_powerup_health_increase)));
			}

			push!(WorldEvent::PlayerEffect { id, effect, time: world.effects.get_time(id, effect) + self.config.effect_time, power });
		}

		events
	}

	fn get_prob(mean_interval: f32, players: usize) -> f32 {
		(DELTA_TIME / mean_interval) * players as f32
	}

	fn find_item_pos(world: &World, rng: &mut WorldRng, config: &EventConfig, radius: f32, should_spawn: impl Fn(Vec2) -> bool) -> Option<Vec2> {
		debug_assert!(!world.players.is_empty()); // Nothing should spawn when there are no players

		for _ in 0..20 { // Gives up if no suitable place is found after too many tries
			// Lazy O(n) algorithm, might improve later
			let rand_player_pos = world.players.iter().nth(rng.next_u32() as usize % world.players.len()).unwrap().1.get_pos();

			// Gets the range for spawning, which must be inside the world
			let min = (rand_player_pos - Vec2::splat(config.item_spawn_radius)).max(radius - world.size / 2.0);
			let max = (rand_player_pos + Vec2::splat(config.item_spawn_radius)).min(world.size / 2.0 - radius);

			let pos = Vec2::new(maths::lerp(min.x, max.x, rng.next_f32()), maths::lerp(min.y, max.y, rng.next_f32()));

			// If not colliding with any block (or some other condition)
			if should_spawn(pos) {
				return Some(pos);
			}
		}

		None
	}
}

impl World {
	fn apply_event(&mut self, event: &WorldEvent, changes: &mut impl Changes) {
		match *event {
			WorldEvent::NewPlayer => self.new_player(),
			WorldEvent::PlayerPos(id, pos) => {
				self.player_present(id, |player| player.set_pos(pos));
				changes.pos_changed(id);
			},
			WorldEvent::PlayerVel(id, vel) => self.player_present(id, |player| player.set_vel(vel)),
			WorldEvent::PlayerHealth(id, health) => self.player_present(id, |player| player.set_health(health)),
			WorldEvent::PlayerAmmo(id, ammo) => self.player_present(id, |player| player.set_ammo_count(ammo)),
			WorldEvent::PlayerEffect { id, effect, time, power } => self.effects.add_effect(id, effect, time, power),
			WorldEvent::NewAmmoCrate(pos) => self.ammo_crates.push(AmmoCrate::new(pos)),
			WorldEvent::NewPowerup(pos, typ) => self.powerups.push(Powerup::new(pos, typ)),
		}
	}

	pub fn apply_events(&mut self, events: &[WorldEvent], changes: &mut impl Changes) {
		for event in events {
			self.apply_event(event, changes);
		}
	}

	fn player_present(&mut self, id: u32, f: impl Fn(&mut Player)) {
		if let Some(player) = self.players.get_mut(&id) {
			f(player);
		}
	}
}
