// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
use glam::Vec2;

use rand::RngCore;

use crate::world::{World, WorldRng};
use crate::player::{Player, RADIUS};
use crate::blocks::Blocks;
use crate::utils::rng::Float;

const FALLBACK_MOVES: usize = 256;
const MAX_GEN_ATTEMPTS: usize = 256;
const MAX_INIT_GEN_ATTEMPTS: usize = FALLBACK_MOVES * MAX_GEN_ATTEMPTS * 4;

pub struct MoveChooser {
	bounds: Vec2,
	fallback: Box<[Vec2]>,
}

impl MoveChooser {
	pub fn build(world: &World, blocks: &Blocks, spawn_radius: f32, rng: &mut WorldRng) -> Result<MoveChooser, String> {
		let bounds = (world.size - RADIUS * 2.0).min(Vec2::splat(spawn_radius));
		let mut moves = Vec::with_capacity(FALLBACK_MOVES);

		let mut i = 0;
		while moves.len() < FALLBACK_MOVES {
			if let Some(mov) = MoveChooser::try_generate(bounds, blocks, rng) {
				moves.push(mov);
			}

			i += 1;
			if i > MAX_INIT_GEN_ATTEMPTS {
				return Err(String::from("fallback move generation taking too long, make sure the player can actually spawn outside of a block near the origin"));
			}
		}

		Ok(MoveChooser { bounds, fallback: moves.into_boxed_slice() })
	}

	pub fn generate(&self, blocks: &Blocks, rng: &mut WorldRng) -> Vec2 {
		for _ in 0..MAX_GEN_ATTEMPTS {
			if let Some(mov) = MoveChooser::try_generate(self.bounds, blocks, rng) {
				return mov;
			}
		}

		self.fallback[rng.next_u32() as usize % self.fallback.len()]
	}

	fn try_generate(bounds: Vec2, blocks: &Blocks, rng: &mut WorldRng) -> Option<Vec2> {
		let pos = (Vec2::new(rng.next_f32(), rng.next_f32()) - 0.5) * bounds;
		(!Player::colliding(pos, blocks)).then_some(pos)
	}
}
