// Copyright Marcus Del Favero 2024
// Licensed under the GNU AGPLv3 with an exception, see `README.md` for details
mod local_player;

pub use local_player::LocalPlayer;

use std::collections::{VecDeque, btree_map::Entry};

use glam::Vec2;

use super::{World, DMap, DELTA_TIME, MAX_STEPS, changes::ClientChanges};

use crate::player::{Player, DPlayer, PlayerController, update::{PlayerUpdate, PlayerBulkUpdate}, forcefield::Forcefield};
use crate::net::{client::ClientError, message::Message};
use crate::playing::model::Model;
use crate::bullet::{Bullet, BulletManager};
use crate::camera::Camera;
use crate::health_bar::{HealthBar, HealthBarRenderer};
use crate::powerup::Effect;
use crate::blocks::Blocks;
use crate::utils::RollingAverage;

pub struct SmoothWorld {
	current_id: u32,
	players: DMap<u32, LocalPlayer>,
	bullets: BulletManager,
	staging_updates: VecDeque<PlayerUpdate>,
	past_staging_updates: RollingAverage,
	update_count: u64,
	time_shift: f32,
}

impl SmoothWorld {
	pub fn new() -> SmoothWorld {
		SmoothWorld {
			current_id: 0,
			players: DMap::new(),
			bullets: BulletManager::new(),
			staging_updates: VecDeque::new(),
			past_staging_updates: RollingAverage::new(256.try_into().unwrap()),
			update_count: 0,
			time_shift: 0.0,
		}
	}

	pub fn reset(&mut self, model: &Model) {
		self.current_id = model.get_id();
		self.players = model.get_world().players.iter().map(|(&id, player)| (id, LocalPlayer::new(player.clone()))).collect();
		self.bullets = model.get_world().bullets.clone();
		self.staging_updates.clear();
		self.past_staging_updates.reset();
		self.update_count = 0;
		self.time_shift = 0.0;
	}

	pub(super) fn get_total_time_shift(&self) -> f32 {
		self.staging_updates.len() as f32 * DELTA_TIME + self.time_shift
	}

	pub fn player_iter(&self) -> impl Iterator<Item = (u32, &LocalPlayer)> {
		self.players.iter().map(|(&id, player)| (id, player))
	}

	pub(super) fn player_iter_mut(&mut self) -> impl Iterator<Item = (u32, &mut LocalPlayer)> {
		self.players.iter_mut().map(|(&id, player)| (id, player))
	}

	pub fn get_player_pos(&self, time_diff: f32) -> Vec2 { self.players[&self.current_id].get_pos(time_diff) }
	pub fn get_player_vel(&self) -> Vec2 { self.players[&self.current_id].get_vel() }
	pub(super) fn get_bullets(&self) -> &[Bullet] { &self.bullets }

	pub fn get_forcefields(&self, world: &World, time_diff: f32) -> Vec<(Forcefield, f32)> {
		let time_shift = self.get_total_time_shift() + time_diff;

		/*
		 * Subtracting time_diff as this time decreases over time rather than
		 * increases (it's the remaining time of the effect).
		 */
		world.effects
			.player_iter_with_effect(Effect::Forcefield)
			.map(|(id, info)| (Forcefield { id, pos: self.players[&id].get_pos(time_diff), strength: info.power }, info.time - time_shift))
			.collect()
	}

	pub fn control_player(&mut self, pcontrol: &PlayerController) {
		self.players.get_mut(&self.current_id).unwrap().player.set_update(pcontrol.get_update());
	}

	pub fn create_bulk_update(&self) -> Message {
		// update_count is the sequence number of the end update
		// Instead want that for the beginning update
		let sequence_number = self.update_count - (self.staging_updates.len() - 1) as u64;
		let mut bulk_update = PlayerBulkUpdate::new(sequence_number);
		for &update in &self.staging_updates {
			bulk_update.push(update);
		}
		Message::PlayerBulkUpdate(bulk_update)
	}

	pub fn sync(&mut self, world: &World, blocks: &Blocks, changes: &ClientChanges, ack: u64, recent_update_count: usize) {
		self.players.retain(|id, _| world.players.contains_key(id));
		debug_assert!(self.players.contains_key(&self.current_id));

		let to_keep = self.update_count.saturating_sub(ack);
		let to_remove = self.staging_updates.len().saturating_sub(to_keep as usize);

		self.time_shift += (to_remove as f32 - recent_update_count as f32) * DELTA_TIME;

		for _ in 0..to_remove { self.staging_updates.pop_front(); }

		self.bullets = world.bullets.clone();

		/*
		 * Extrapolates the current world's state by the length of the staging
		 * update queue.
		 *
		 * For the current player this is applying the staging updates, and for
		 * all other players this is using the most recent update, as an
		 * estimation that likely will be the next one.
		 *
		 * Also updating the bullets into the future while doing this.
		 */
		let mut new_players = world.players.clone();
		for &update in &self.staging_updates {
			for (&id, player) in &mut new_players {
				if id == self.current_id { // Use the most recent update (which was already set) as a prediction if not the current player
					player.set_update(update);
				}

				SmoothWorld::update_player(&mut self.bullets, world, blocks, player, id, id == self.current_id);
			}

			SmoothWorld::update_bullets(&mut self.bullets, DELTA_TIME, world, blocks, new_players.iter_mut());
		}
		SmoothWorld::update_bullets(&mut self.bullets, self.time_shift, world, blocks, new_players.iter_mut());

		for (&id, player) in &new_players {
			match self.players.entry(id) {
				Entry::Vacant(e) => _ = e.insert(LocalPlayer::new(player.clone())),
				Entry::Occupied(mut e) => {
					let prev_player = e.get_mut();

					if !changes.moved.contains(&id) {
						let delta = &prev_player.player - player;
						prev_player.delta += &delta;
						if id == self.current_id {
							prev_player.delta.reset_angle();
						}
					} else {
						prev_player.delta = DPlayer::zero();
					}

					prev_player.player = player.clone();
				}
			}
		}

		for hit in &changes.hits {
			let player = self.players.get_mut(&hit.player_id).unwrap();
			player.last_hurt = if hit.killed { f32::INFINITY } else { 0.0 };
			if hit.killed {
				player.hbar.reset();
			}
		}
	}

	pub fn update(&mut self, update: PlayerUpdate, world: &World, blocks: &Blocks) -> Result<(), ClientError> {
		self.players.get_mut(&self.current_id).unwrap().player.set_update(update);

		if self.staging_updates.len() >= MAX_STEPS {
			return Err(ClientError::Limit(String::from("too many staging updates accumulated by client")));
		}

		self.staging_updates.push_back(update);
		self.update_count += 1;

		self.past_staging_updates.add(self.staging_updates.len() as f32);
		let mean_len = self.past_staging_updates.mean().unwrap();
		let decay_rate = ((100.0 / (mean_len + 1.0)) + 1.0) * DELTA_TIME;

		let prev_time_shift = self.time_shift;
		self.time_shift *= (1.0 - DELTA_TIME * 2.0).max(0.0);

		for (&id, player) in &mut self.players {
			player.delta.decay(decay_rate);
			SmoothWorld::update_player(&mut self.bullets, world, blocks, &mut player.player, id, id == self.current_id);
		}

		SmoothWorld::update_bullets(&mut self.bullets, DELTA_TIME - prev_time_shift + self.time_shift, world, blocks, self.players.iter_mut().map(|(id, player)| (id, &mut player.player)));

		Ok(())
	}

	fn update_player(bullets: &mut BulletManager, world: &World, blocks: &Blocks, player: &mut Player, id: u32, is_current: bool) {
		player.update(DELTA_TIME, world.size, blocks, &world.effects, id);
		if is_current {
			player.decrease_cooldown(DELTA_TIME * world.effects.get_mul(id, Effect::Reload));
			if player.shoot() {
				bullets.add(Bullet::new(player, id, world.effects.get_mul(id, Effect::Damage), 0.0));
			}
		}
	}

	fn update_bullets<'a>(bullets: &mut BulletManager, dt: f32, world: &World, blocks: &Blocks, players: impl Iterator<Item = (&'a u32, &'a mut Player)>) {
		bullets.update(dt, &world.get_forcefields());
		for (&id, player) in players {
			for bullet in bullets.collide_player(player, id) {
				player.on_bullet_hit(&bullet);
			}
		}
		bullets.collide_blocks(blocks);
	}

	pub fn push_health_bars(&self, hbar_renderer: &mut HealthBarRenderer, lwinsize: Vec2, camera: &Camera, time_diff: f32) {
		// Render this player's health bar on top of all others
		let builder = HealthBar::above_player_builder(lwinsize, camera);
		for (&id, player) in &self.players {
			if id != self.current_id {
				hbar_renderer.push(builder.make(player, &player.hbar, time_diff));
			}
		}

		hbar_renderer.push(self.players[&self.current_id].hbar.make_hud(lwinsize));
	}
}
